;	========================================
;	Motorsteller (Fahrtregler) mit AT90S2313
;	========================================
;
;	Version 1.03 (17.05.2001)
;
;	System arbeitet mit 4,0 MHz
;	Timer 1 luft mit Vorteiler=8 (Takt=500kHz)
;
;	PortB-Belegung:	Bit 0 -	Ausgang unten links   - Ruhezustand=1
;			Bit 1 - Ausgang unten rechts  - Ruhezustand=1
;			Bit 2 - Ausgang oben links    - Ruhezustand=0
;			Bit 3 - Ausgang oben rechts   - Ruhezustand=0
;			Bit 4 - Ausgang Brems-LED     - Ruhezustand=1
;			Bit 5 - Eingang ungenutzt     - Pullup=1
;			Bit 6 - Eingang ungenutzt     - Pullup=1
;			Bit 7 - Eingang ungenutzt     - Pullup=1
;
;	PortD-Belegung:	Bit 0 -	Eingang vom Empfnger - Pullup=0
;			Bit 1 - Eingang ungenutzt     - Pullup=1
;			Bit 2 - Eingang ungenutzt     - Pullup=1
;			Bit 3 - Eingang ungenutzt     - Pullup=1
;			Bit 4 - Eingang Schalter 1    - Pullup=1
;			Bit 5 - Eingang Schalter 2    - Pullup=1
;			Bit 6 - Eingang Schalter 3    - Pullup=1
;			Bit 7 - Eingang ungenutzt     - Pullup=1
;
;	Schalter 1: aus - volle Leistung
;		    ein - halbe Leistung
;	Schalter 2: aus - Leerlauf bei Mittelstellung
;		    ein - Bremse bei Mittelstellung
;	Schalter 3: aus - Normalbetrieb
;		    ein - Lernmodus (Mittelstellung ermitteln und
;			  speichern)
;
;	Speicheradressen
;
.equ	ctstat=	0x60		;Status des Steuereinganges
.equ	pwmmod=	ctstat+1	;PWM-Mode (0=aus,1=vor,2=rck,3=stop)
.equ	pwmcnl=	pwmmod+1	;Wert (L) fr PWM-Steuerung
.equ	pwmcnh=	pwmcnl+1	;Wert (H) fr PWM-Steuerung
.equ	pwmmo2=	pwmcnh+1	;PWM-Mode fr nachsten PWM-Zyklus
.equ	pwmc2l=	pwmmo2+1	;Wert (L) fr nchsten PWM-Zyklus
.equ	pwmc2h=	pwmc2l+1	;Wert (H) fr nchsten PWM-Zyklus
.equ	swstat=	pwmc2h+1	;Status der DIP-Schalter
				;Bit0: 0=halbe Leistung (50%)
				;Bit1: 0=Bremse ein
				;Bit2: 0=Lernmodus
.equ	center=	swstat+1	;Mittelstellung,    2 Byte (~1,50ms)
.equ	forerr=	center+2	;Vorwrts-Fehler,   2 Byte (-0,70ms)
.equ	formin=	forerr+2	;Vorwrts-Minimum,  2 Byte (-0,05ms)
.equ	bckmin=	formin+2	;Rckwrts-Minimum, 2 Byte (+0,05ms)
.equ	bckerr=	bckmin+2	;Rckwrts-Fehler,  2 Byte (+0,70ms)
.equ	lerntb=	bckerr+2	;Wertetabelle fr Lernmodus (8x2 Byte)
;
;	Konstanten
;
.equ	kcentr=	750		;Normalwert Mittelstellung  (~1,50ms)
.equ	kcrmin=	kcentr-100	;Minimalwert Mittelstellung (-0,20ms)
.equ	kcrmax=	kcentr+100	;Maximalwert Mittelstellung (+0,20ms)
;
.equ	kfoerr=	-350		;Offset Vorwrts-Fehler
.equ	kfomin=	-25		;Offset Vorwrts-Minimum
.equ	kbkmin=	25		;Offset Rckwrts-Minimum
.equ	kbkerr=	350		;Offset Rckwrts-Fehler
;
.equ	stndby=	0b11110011	;Ruhezustand - alle Treiber aus
.equ	forwrd=	0b11110101	;Vorwrts    - Treiber 3+2 aktiv
.equ	bckwrd=	0b11111010	;Rckwrts   - Treiber 1+4 aktiv
.equ	brake=	0b11110000	;Bremse      - Treiber 1+2 aktiv
.equ	led=	0b11101111	;LED (wird mittels AND mit Treiber-
				;Status verknpft wenn aktiv)
;
.equ	pulsee=	15000		;Impulsabstand-Fehler (>30ms)
;
.include "2313def.inc"
.cseg
.org 0
;
reset:	rjmp	start			;Programm-Beginn
exint0:	reti				;wird nicht genutzt
exint1:	reti				;wird nicht genutzt
tcapt1:	reti				;wird nicht genutzt
tcomp1:	reti				;wird nicht genutzt
tovf1:	reti				;wird nicht genutzt
tovf0:	reti				;wird nicht genutzt
uarecx:	reti				;wird nicht genutzt
uadre:	reti				;wird nicht genutzt
uatxc:	reti				;wird nicht genutzt
ancomp:	reti				;wird nicht genutzt
;
start:	ldi	r16,low(ramend)		;Stackpointer auf RAM-Ende
	out	spl,r16			;setzen
	ldi	r16,stndby&led		;PortB, Ausgnge/Pullups
	out	portb,r16		;setzen
	ldi	r16,0b11111110		;PortD, Pullups
	out	portd,r16		;setzen
	ldi	r16,0b00011111		;PortB, Eingnge/Ausgnge
	out	ddrb,r16		;setzen
	clr	r16			;PortD, Eingnge
	out	ddrd,r16		;setzen
	out	tccr1a,r16		;Timer1, OC/PWM ausschalten
	ldi	r16,1<<cs11		;Timer1, Taktquelle auf Clk/8
	out	tccr1b,r16		;setzen und starten
	ldi	xl,0x60			;Zeiger auf RAM-Anfang
	clr	xh
ramclr:	st	x+,xh			;Speicherzelle lschen
	cpi	xl,ramend+1		;Speicherende erreicht?
	brne	ramclr			;nein -> weiter lschen
;
	in	r16,pind		;Schalter-Eingnge lesen
	swap	r16			;Bits 4-6 auf 0-2 setzen
	andi	r16,0x07		;Schalter-Bits filtern
	sts	swstat,r16		;und speichern
	sbrs	r16,2			;Lernmodus aktiviert?
	rjmp	lern			;ja -> Sprung zum Lernprogramm
;
	ldi	r17,centrl		;EEPROM-Adresse Mittelwert L
	rcall	readee			;EEPROM lesen
	sts	center,r18		;Mittelwert L speichern
	inc	r17			;EEPROM-Adresse Mittelwert H
	rcall	readee			;EEPROM lesen
	sts	center+1,r18		;Mittelwert H speichern
;
	lds	r16,center		;Mittelwert L holen
	lds	r17,center+1		;Mittelwert H holen
	ldi	r18,low(kcrmax)		;Maximalwert L laden
	ldi	r19,high(kcrmax)	;Maximalwert H laden
	sub	r18,r16
	sbc	r19,r17			;Maximalwert berschritten?
	brcs	error			;ja -> Fehleranzeige
	ldi	r18,low(kcrmin)		;Minimalwert L laden
	ldi	r19,high(kcrmin)	;Minimalwert H laden
	sub	r16,r18
	sbc	r17,r19			;Minimalwert unterschritten?
	brcc	calc			;nein -> Vergleichswerte berech
;
error:	ldi	r19,stndby&led		;Treiber auf Standby, LED ein
	out	portb,r19		;ausgeben
	ldi	r18,3			;Warteschleife ca. 150ms
	rcall	delay
	ldi	r19,stndby		;Treiber auf Standby, LED aus
	out	portb,r19		;ausgeben
	ldi	r18,3			;Warteschleife ca. 150ms
	rcall	delay
	rjmp	error			;wiederholen (LED blinkt)
;
calc:	lds	r16,center		;Mittelwert L holen
	lds	r17,center+1		;Mittelwert H holen
	ldi	r18,low(kfoerr)		;Offset Vorwrts-Fehler L
	ldi	r19,high(kfoerr)	;Offset Vorwrts-Fehler H
	add	r18,r16			;Vorwrts-Fehler-Wert
	adc	r19,r17			;berechnen und
	sts	forerr,r18		;Ergebnis L speichern
	sts	forerr+1,r19		;Ergebnis H speichern
	ldi	r18,low(kfomin)		;Offset Vorwrts-Minimum L
	ldi	r19,high(kfomin)	;Offset Vorwrts-Minimum H
	add	r18,r16			;Vorwrts-Minimum-Wert
	adc	r19,r17			;berechnen und
	sts	formin,r18		;Ergebnis L speichern
	sts	formin+1,r19		;Ergebnis H speichern
	ldi	r18,low(kbkmin)		;Offset Rckwrts-Minimum L
	ldi	r19,high(kbkmin)	;Offset Rckwrts-Minimum H
	add	r18,r16			;Rckwrts-Minimum-Wert
	adc	r19,r17			;berechnen und
	sts	bckmin,r18		;Ergebnis L speichern
	sts	bckmin+1,r19		;Ergebnis H speichern
	ldi	r18,low(kbkerr)		;Offset Rckwrts-Fehler L
	ldi	r19,high(kbkerr)	;Offset Rckwrts-Fehler H
	add	r18,r16			;Rckwrts-Maximum-Wert
	adc	r19,r17			;berechnen und
	sts	bckerr,r18		;Ergebnis L speichern
	sts	bckerr+1,r19		;Ergebnis H speichern
;	
	ldi	r18,20			;ca. 1 Sekunde warten auf
	rcall	delay			;stabilen Datenempfang
;
;	Neustart des Datenempfangs
;
restrt:	ldi	r16,stndby		;Treiber auf Standby
	out	portb,r16		;setzen und LED ausschalten
	clr	r16
	sts	pwmmod,r16		;PWM ausschalten
	sts	pwmcnl,r16		;PWM-Wert lschen
	sts	pwmcnh,r16
	sts	pwmmo2,r16		;nchste PWM ausschalten
	sts	pwmc2l,r16		;nchsten PWM-Wert lschen
	sts	pwmc2h,r16
	out	tcnt1h,r16		;Timer1, Zhler H rcksetzen
	out	tcnt1l,r16		;Timer1, Zhler L rcksetzen
;
;	-------------
;	Hauptschleife
;	-------------
;
main:	in	r16,pind		;Dateneingang lesen
	andi	r16,0x01		;Dateneingang filtern
	lds	r17,ctstat		;letzten Status lesen
	cp	r16,r17			;Status-nderung?
	sts	ctstat,r16		;neuen Status speichern
	brcs	mai200			;Impulsende -> weiter
	brne	mai100			;Impulsanfang -> weiter
	rjmp	mai300			;keine nderung -> weiter
;
;	Impulsbeginn; Zhler starten
;
mai100:	clr	r16
	out	tcnt1h,r16		;Timer1, Zhler H rcksetzen
	out	tcnt1l,r16		;Timer1, Zhler L rcksetzen
	lds	r16,pwmc2l		;ermittelten PWM-Wert holen
	lds	r17,pwmc2h
	sts	pwmcnl,r16		;und neu setzen
	sts	pwmcnh,r17
	lds	r16,pwmmo2		;ermittelten PWM-Mode holen
	sts	pwmmod,r16		;und neu setzen
	ldi	r16,stndby		;Treiber zunchst auf Standby
	out	portb,r16
	lds	r16,pwmmod		;PWM-Modus holen
	tst	r16			;PWM aus?
	brne	mai110			;nein -> weiter
	ldi	r17,stndby		;Treiber auf Standby
	rjmp	mai150
mai110:	dec	r16			;PWM vorwrts?
	brne	mai120			;nein -> weiter
	ldi	r17,forwrd		;Treiber auf Vorwrts
	rjmp	mai150
mai120:	dec	r16			;PWM rckwrts?
	brne	mai130			;nein -> weiter
	ldi	r17,bckwrd		;Treiber auf Rckwrts
	rjmp	mai150
mai130:	lds	r16,swstat		;Schalter-Status holen
	bst	r16,1			;Schalter auf "Bremse ein"?
	brts	mai140			;ja -> weiter mit Standby
	ldi	r17,brake&led		;Treiber auf Bremse + LED
	rjmp	mai150
mai140:	ldi	r17,stndby&led		;Treiber auf Standby + LED
mai150:	out	portb,r17		;Treiber neu setzen
	rjmp	main
;
;	Impulsende; Zhler lesen
;
mai200:	in	r16,tcnt1l		;Timer1, Zhler lesen
	in	r17,tcnt1h
	lds	r18,bckerr		;Fehlerschwelle rckwrts
	lds	r19,bckerr+1
	sub	r18,r16			;Fehlerschwelle berschritten?
	sbc	r19,r17
	brcc	mai210			;nein -> weiter
	rjmp	maierr			;sonst -> Fehler
mai210:	lds	r18,bckmin		;Rckwrts Minimum
	lds	r19,bckmin+1
	sub	r18,r16			;Rckwrts Min berschritten?
	sbc	r19,r17
	brcs	maibck			;ja -> Rckwrts
	lds	r18,formin		;Vorwrts Minimum
	lds	r19,formin+1
	sub	r18,r16			;Vorwrts Min berschritten?
	sbc	r19,r17
	brcs	maistp			;ja -> Stop/Bremse
	lds	r18,forerr		;Fehlerschwelle vorwrts
	lds	r19,forerr+1
	mov	r20,r16			;Zhler kopieren
	mov	r21,r17
	sub	r20,r18			;Fehlerschwelle berschritten?
	sbc	r21,r19
	brcs	maierr			;ja -> Fehler
;
maifor:	lds	r18,formin		;Vorwrts-Minimal-Wert
	lds	r19,formin+1
	sub	r18,r16			;Differenz berechnen
	sbc	r19,r17
	ldi	r20,6			;6x verdoppeln (*64)
	lds	r21,swstat		;Schalter-Status holen
	bst	r21,0			;Schalter "halbe Leistung"?
	brts	maif10			;nein -> weiter
	dec	r20			;sonst nur 5x verdoppeln (*32)
maif10:	clc				;Carry lschen
	rol	r18			;Wert verdoppeln
	rol	r19
	dec	r20			;Wert 6x verdoppelt?
	brne	maif10			;nein -> Schleife
	sts	pwmc2l,r18		;PWM-Wert L vorbereiten
	sts	pwmc2h,r19		;PWM-Wert H vorbereiten
	ldi	r20,1
	sts	pwmmo2,r20		;PWM auf Vorwrts setzen
	rjmp	main
;
maibck:	lds	r18,bckmin		;Rckwrts-Minimal-Wert
	lds	r19,bckmin+1
	sub	r16,r18			;Differenz berechnen
	sbc	r17,r19
	ldi	r20,6			;6x verdoppeln (*64)
	lds	r21,swstat		;Schalter-Status holen
	bst	r21,0			;Schalter "halbe Leistung"?
	brts	maib10			;nein -> weiter
	dec	r20			;sonst nur 5x verdoppeln (*32)
maib10:	clc				;Carry lschen
	rol	r16			;Wert verdoppeln
	rol	r17
	dec	r20			;Wert 6x verdoppelt?
	brne	maib10			;nein -> Schleife
	sts	pwmc2l,r16		;PWM-Wert L vorbereiten
	sts	pwmc2h,r17		;PWM-Wert H vorbereiten
	ldi	r20,2
	sts	pwmmo2,r20		;PWM auf Rckwrts setzen
	rjmp	main
;
maistp:	ldi	r20,3
	sts	pwmmo2,r20		;PWM auf Stop/Bremse setzen
	clr	r20
	sts	pwmc2l,r20		;PWM-Wert lschen
	sts	pwmc2h,r20
	rjmp	main
;
maierr:	clr	r20
	sts	pwmmo2,r20		;PWM ausschalten
	sts	pwmc2l,r20		;PWM-Wert lschen
	sts	pwmc2h,r20
	rjmp	main
;
mai300:	in	r16,tcnt1l		;Timer1, Zhler lesen
	in	r17,tcnt1h
	ldi	r18,low(pulsee)		;Pulsabstands-Fehlerwert
	ldi	r19,high(pulsee)
	sub	r18,r16			;Fehlerschwelle berschritten?
	sbc	r19,r17
	brcs	mai350			;ja -> Neustart
	lds	r20,pwmmod		;PWM-Modus holen
	tst	r20			;PWM ausgeschaltet?
	brne	mai310			;nein -> weiter
	rjmp	main			;sonst Ende
;
mai310:	cpi	r20,3			;PWM auf Stop gesetzt?
	brne	mai320			;nein -> weiter
	rjmp	main			;sonst Ende
;
mai320:	lds	r18,pwmcnl		;PWM-Wert holen
	lds	r19,pwmcnh
	sub	r18,r16			;Ausschaltschwelle erreicht?
	sbc	r19,r17
	brcs	mai330			;ja -> weiter
	rjmp	main
;
mai330:	ldi	r16,stndby		;PWM in Ruhephase
	out	portb,r16		;ausgeben
	rjmp	main
;
mai350:	rjmp	restrt			;Datenfehler -> Neustart
;
;
;	Lernmodus, 8 Impulse einlesen und Mittelwert bilden
;	wenn ok, dann in EEPROM speichern
;
lern:	ldi	r18,20			;ca. 1 Sekunde warten auf
	rcall	delay			;stabilen Datenempfang
	ldi	r16,stndby		;Treiber auf Standby, LED aus
	out	portb,r16		;ausgeben
	clr	r22			;Zykluszhler zurcksetzen
	out	tcnt1h,r22		;Timer1, Zhler H rcksetzen
	out	tcnt1l,r22		;Timer1, Zhler L rcksetzen
	ldi	xl,low(lerntb)		;Zeiger auf Wertetabelle
	ldi	xh,high(lerntb)
ler010:	in	r16,pind		;Dateneingang lesen
	andi	r16,0x01		;Dateneingang filtern
	lds	r17,ctstat		;letzten Status lesen
	cp	r16,r17			;Status-nderung?
	sts	ctstat,r16		;neuen Status speichern
	brcs	ler200			;Impulsende -> weiter
	brne	ler100			;Impulsanfang -> weiter
	rjmp	ler300			;keine nderung -> weiter
;
;	Impulsbeginn; Zhler starten
;
ler100:	clr	r16
	out	tcnt1h,r16		;Timer1, Zhler H rcksetzen
	out	tcnt1l,r16		;Timer1, Zhler L rcksetzen
	rjmp	ler010
;
ler200:	in	r16,tcnt1l		;Timer1, Zhler lesen
	in	r17,tcnt1h
	ldi	r18,low(kcrmax)		;Maximalwert L laden
	ldi	r19,high(kcrmax)	;Maximalwert H laden
	sub	r18,r16
	sbc	r19,r17			;Maximalwert berschritten?
	brcc	ler210			;nein -> weiter
	rjmp	error			;sonst Fehleranzeige
;
ler210:	mov	r20,r16			;gelesenen Timer-Wert
	mov	r21,r17			;sichern
	ldi	r18,low(kcrmin)		;Minimalwert L laden
	ldi	r19,high(kcrmin)	;Minimalwert H laden
	sub	r16,r18
	sbc	r17,r19			;Minimalwert unterschritten?
	brcc	ler220			;nein -> Wert speichern
	rjmp	error			;sonst Fehleranzeige
;
ler220:	st	x+,r20			;Wert speichern L
	st	x+,r21			;Wert speichern H
	inc	r22			;Zykluszhler erhhen
	cpi	r22,8			;alle Zyklen komplett?
	brcs	ler010			;nein -> weiter
;	
	ldi	xl,low(lerntb)		;Zeiger auf Wertetabelle
	ldi	xh,high(lerntb)
	ldi	r22,7			;7 Additionen
	ld	r19,x+			;ersten L-Wert holen
	ld	r20,x+			;ersten H-Wert holen
	clr	r21			;bertrag lschen
	clr	r18			;bertrag lschen
ler230:	ld	r16,x+			;nchten L-Wert holen
	ld	r17,x+			;nchten H-Wert holen
	add	r19,r16			;L-Wert addieren
	adc	r20,r17			;H-Wert addieren
	adc	r21,r18			;bertrag addieren
	dec	r22			;alle Werte addiert?
	brne	ler230			;nein -> Schleife
	ror	r21			;Ergebnis durch 8 dividieren
	ror	r20			;(runden)
	ror	r19			;Ergebnis /2
	clc
	ror	r21
	ror	r20
	ror	r19			;Ergebnis /2
	clc
	ror	r21
	ror	r20
	ror	r19			;Ergebnis /2
	mov	r18,r19			;Endwert L in EEPROM
	ldi	r17,centrl		;EEPROM-Adresse setzen
	rcall	writee			;schreiben
	mov	r18,r20			;Endwert H in EEPROM
	ldi	r17,centrh		;EEPROM-Adresse setzen
	rcall	writee			;schreiben
	ldi	r16,stndby&led		;sonst Standby und LED ein
	out	portb,r16		;ausgeben
ler240:	rjmp	ler240			;Ende
;
ler300:	in	r16,tcnt1l		;Timer1, Zhler lesen
	in	r17,tcnt1h
	ldi	r18,low(pulsee)		;Pulsabstands-Fehlerwert
	ldi	r19,high(pulsee)
	sub	r18,r16			;Fehlerschwelle berschritten?
	sbc	r19,r17
	brcc	ler310			;nein -> weiter
	rjmp	error			;sonst Fehleranzeige
ler310:	rjmp	ler010
;
;	Unterprogramm zum Lesen des EEPROM
;	r16 - Zwischenspeicher
;	r17 - EEPROM-Adresse
;	r18 - EEPROM-Daten
;
readee:	sbic	eecr,eewe		;Schreibvorgang noch aktiv?
	rjmp	readee			;ja -> Schleife
	out	eear,r17		;EEPROM-Adresse setzen
	sbi	eecr,eere		;EEPROM-Lese-Bit setzen
	in	r18,eedr		;EEPROM-Byte auslesen
	clr	r16
	out	eear,r16		;EEPROM-Adresse=0 setzen
	ret
;
;	Unterprogramm zum Schreiben in den EEPROM
;	r16 - Zwischenspeicher
;	r17 - EEPROM-Adresse
;	r18 - EEPROM-Daten
;
writee:	sbic	eecr,eewe		;luft Schreibzyklus?
	rjmp	writee			;ja -> warten
	out	eear,r17		;EEPROM-Adresse setzen
	out	eedr,r18		;EEPROM-Datenbyte setzen
	sbi	eecr,eemwe		;Masterwrite-Bit setzen
	sbi	eecr,eewe		;Write-Bit setzen
;
wri010:	sbic	eecr,eewe		;Schreibvorgang noch aktiv?
	rjmp	wri010			;ja -> Schleife
	clr	r16
	out	eear,r16		;EEPROM-Adresse=0 setzen
	ret
;
;	Unterprogramm Warteschleife (ca. 50ms)
;	r18 - Anzahl der Zyklen
;
delay:	clr	r17			;M-Zhler lschen
	clr	r16			;L-Zhler lschen
del010:	dec	r16			;L-Zhler bearbeiten
	brne	del010			;Schleife
	dec	r17			;M-Zhler bearbeiten
	brne	del010			;Schleife
	dec	r18			;H-Zhler bearbeiten
	brne	del010			;Schleife
	ret
;
;	EEPROM-Inhalt
;
.eseg
;
	.db	0xff			;Byte 0 wird nicht genutzt
versnr:	.db	1,0,3			;Versions-Nummer x.xx
	.db	0xff			;nicht genutzt
centrl:	.db	low(kcentr)		;Wert bei Mittelstellung (L)
centrh:	.db	high(kcentr)		;Wert bei Mittelstellung (H)
